#ifndef _DYN_MANAGER_H
#define _DYN_MANAGER_H

#ifdef __cplusplus
extern "C" {
#endif

#include "dyn_manager_registers.h"
#include "dynamixel_master.h"
#include "memory.h"
#include "scheduler.h"

#ifndef DYN_MANAGER_MAX_NUM_MASTERS
  #error "Please, specify the maximum number of masters with the DYN_MANAGER_MAX_NUM_MASTERS macro"
#endif
#ifndef DYN_MANAGER_MAX_NUM_MODULES
  #error "Please, specify the maximum number of modules with the DYN_MANAGER_MAX_NUM_MODULES macro"
#endif
#ifndef DYN_MANAGER_MAX_NUM_DEVICES
  #error "Please, specify the maximum number of devices with the DYN_MANAGER_MAX_NUM_DEVICES macro"
#endif
#ifndef DYN_MANAGER_MAX_NUM_SINGLE_OP
  #error "Please, specify the maximum number of single operations with the DYN_MANAGER_MAX_NUM_SINGLE_OP macro"
#endif
#ifndef DYN_MANAGER_MAX_NUM_SYNC_OP
  #error "Please, specify the maximum number of sync operations with the DYN_MANAGER_MAX_NUM_SYNC_OP macro"
#endif
#ifndef DYN_MANAGER_MAX_NUM_BULK_OP
  #error "Please, specify the maximum number of bulk operations with the DYN_MANAGER_MAX_NUM_BULK_OP macro"
#endif
#define    DYN_MANAGER_MAX_NUM_OP            ((DYN_MANAGER_MAX_NUM_SINGLE_OP+DYN_MANAGER_MAX_NUM_SYNC_OP+DYN_MANAGER_MAX_NUM_BULK_OP)*DYN_MANAGER_MAX_NUM_MASTERS)

#define    DYN_MANAGER_PERIODIC              0xFF

#define    DYN_MANAGER_MODE_MASK             0x0F
#define    DYN_MANAGER_READ                  0x00
#define    DYN_MANAGER_WRITE                 0x01

#define    DYN_MANAGER_OP_MASK               0xF0
#define    DYN_MANAGER_SINGLE                0x10
#define    DYN_MANAGER_SYNC                  0x20
#define    DYN_MANAGER_BULK                  0x40
#define    DYN_MANAGER_NONE                  0x00

typedef enum {single_rd_op=DYN_MANAGER_SINGLE|DYN_MANAGER_READ,
              single_wr_op=DYN_MANAGER_SINGLE|DYN_MANAGER_WRITE,
              sync_rd_op=DYN_MANAGER_SYNC|DYN_MANAGER_READ,
              sync_wr_op=DYN_MANAGER_SYNC|DYN_MANAGER_WRITE,
              bulk_rd_op=DYN_MANAGER_BULK|DYN_MANAGER_READ,
              bulk_wr_op=DYN_MANAGER_BULK|DYN_MANAGER_WRITE,
              no_op=DYN_MANAGER_NONE} op_type_t;

typedef struct{
  op_type_t op_type;
  unsigned char op_index[DYN_MANAGER_MAX_NUM_MASTERS];
  unsigned char handle_index;
}OP_HANDLE;

typedef struct{
  op_type_t operation;
  unsigned char repetitions;
  unsigned char period_count;
  OP_HANDLE *op_handle;
}TDynManagerOpCommon;

typedef struct{
  TDynManagerOpCommon common;
  unsigned char id;
  unsigned short int address;
  unsigned short int length;
  unsigned char *data;
}TDynManagerSingleOp;

typedef struct{
  TDynManagerOpCommon common;
  unsigned char num_devices;
  unsigned char ids[DYN_MANAGER_MAX_NUM_DEVICES];
  unsigned short int address;
  unsigned short int length;
  unsigned char *data[DYN_MANAGER_MAX_NUM_DEVICES];
}TDynManagerSyncOp;

typedef struct{
  TDynManagerOpCommon common;
  unsigned char num_devices;
  unsigned char ids[DYN_MANAGER_MAX_NUM_DEVICES];
  unsigned short int address[DYN_MANAGER_MAX_NUM_DEVICES];
  unsigned short int length[DYN_MANAGER_MAX_NUM_DEVICES];
  unsigned char *data[DYN_MANAGER_MAX_NUM_DEVICES];
}TDynManagerBulkOp;

typedef struct{
  TDynManagerSingleOp single_op[DYN_MANAGER_MAX_NUM_SINGLE_OP];
  unsigned char num_single_op;
  TDynManagerSyncOp sync_op[DYN_MANAGER_MAX_NUM_SYNC_OP];
  unsigned char num_sync_op;
  TDynManagerBulkOp bulk_op[DYN_MANAGER_MAX_NUM_BULK_OP];
  unsigned char num_bulk_op;
}TDynManagerOps;

typedef struct{
  unsigned short model;
  unsigned char master_index;
}TDynManagerDevice;

struct TDynModule;

typedef struct{
  unsigned short int eeprom_base_address;
  unsigned short int ram_base_address;
  TDynamixelMaster *masters[DYN_MANAGER_MAX_NUM_MASTERS];
  TDynManagerOps operations[DYN_MANAGER_MAX_NUM_MASTERS];
  unsigned char num_masters;
  struct TDynModule *modules[DYN_MANAGER_MAX_NUM_MODULES];
  unsigned char num_modules;
  TDynManagerDevice devices[DYN_MANAGER_MAX_NUM_DEVICES];
  unsigned char num_devices;
  OP_HANDLE op_handles[DYN_MANAGER_MAX_NUM_OP];
  unsigned char num_ops;
  unsigned char period_ms;
  TScheduler *scheduler;
  sched_channel_t sch_channel;
  TMemory *memory;
  TMemModule mem_module;
  unsigned char running;
  unsigned char do_scan;
  unsigned int present_devices;
}TDynManager;

// public functions
unsigned char dyn_manager_init(TDynManager *manager,TMemory *memory,TScheduler *scheduler,sched_channel_t sch_channel,unsigned short int eeprom_base_address,unsigned short int ram_base_address);
void dyn_manager_set_period(TDynManager *manager,unsigned char period_ms);
static inline unsigned char dyn_manager_get_period(TDynManager *manager)
{
  return manager->period_ms;
}
void dyn_manager_start(TDynManager *manager);
void dyn_manager_stop(TDynManager *manager);
void dyn_manager_scan(TDynManager *manager);
void dyn_manager_add_master(TDynManager *manager,TDynamixelMaster *master);
static inline unsigned char dyn_manager_get_num_masters(TDynManager *manager)
{
  return manager->num_masters;
}
TDynamixelMaster *dyn_manager_get_device_master(TDynManager *manager,unsigned char device_id);
static inline unsigned short int dyn_manager_get_device_model(TDynManager *manager,unsigned char device_id)
{
  return manager->devices[device_id].model;
}
void dyn_manager_add_module(TDynManager *manager,struct TDynModule *module);
static inline unsigned char dyn_manager_get_num_modules(TDynManager *manager)
{
  return manager->num_modules;
}
void dyn_manager_delete_op(TDynManager *manager,OP_HANDLE *op);
void dyn_manager_set_op_period(TDynManager *manager,OP_HANDLE *op,unsigned char period_count);
unsigned char dyn_manager_get_op_period(TDynManager *manager,OP_HANDLE *op);
void dyn_manager_set_op_repetitions(TDynManager *manager,OP_HANDLE *op,unsigned char repetitions);
unsigned char dyn_manager_get_op_repetitions(TDynManager *manager,OP_HANDLE *op);
// single operation functions
OP_HANDLE *dyn_manager_single_op_new(TDynManager *manager,unsigned char mode,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data);
void dyn_manager_single_op_change_id(TDynManager *manager,OP_HANDLE *op,unsigned char id);
void dyn_manager_single_op_change_address(TDynManager *manager,OP_HANDLE *op,unsigned short int address);
void dyn_manager_single_op_change_length(TDynManager *manager,OP_HANDLE *op,unsigned short int length);
void dyn_manager_single_op_change_data(TDynManager *manager,OP_HANDLE *op,unsigned char *data);
// sync operation functions
OP_HANDLE *dyn_manager_sync_op_new(TDynManager *manager,unsigned char mode,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[]);
void dyn_manager_sync_op_add_devices(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned char * const data[]);
void dyn_manager_sync_op_remove_devices(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids);
void dyn_manager_sync_op_change_address(TDynManager *manager,OP_HANDLE *op,unsigned short int address);
void dyn_manager_sync_op_change_length(TDynManager *manager,OP_HANDLE *op,unsigned short int length);
void dyn_manager_sync_op_change_data(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned char * const data[]);
// bulk operation functions
OP_HANDLE *dyn_manager_bulk_op_new(TDynManager *manager,unsigned char mode,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[]);
void dyn_manager_bulk_op_add_devices(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length,unsigned char * const data[]);
void dyn_manager_bulk_op_remove_devices(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids);
void dyn_manager_bulk_op_change_address(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned short int *address);
void dyn_manager_bulk_op_change_length(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned short int *length);
void dyn_manager_bulk_op_change_data(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned char * const data[]);
void dyn_manager_loop(TDynManager *manager);

#ifdef __cplusplus
  }
#endif

#endif 
